/*****************************************************************

Copyright (C) 2006,2009 Lubos Lunak <l.lunak@kde.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#include "wmiface.cpp"

#include <qtextstream.h>

static void usage()
    {
    QTextStream( stderr ) << "Unknown or incorrect command." << endl
        << "See the README for documentation." << endl;
    exit( 1 );
    }

static QString version()
    {
    return WMIFACE_VERSION_STRING;
    }

// all arg() are based on dcop code, to make it compatible with KDE3
template< typename T >
static T arg( const char* arg );

template<>
bool arg( const char* arg )
    {
    QString s = QString::fromLocal8Bit( arg );
    if ( s.toLower()  == "true" )
        return true;
    if ( s.toLower()  == "yes" )
        return true;
    if ( s.toLower()  == "on" )
        return true;
    if ( s.toInt() != 0 )
        return true;
    return false;
    }

template<>
QPoint arg( const char* arg )
    {
    const char *s = arg;
    char *end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int x = strtol(s, &end, 10);
    s = (const char *)end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int y = strtol(s, &end, 10);
    return QPoint( x, y );
    }

template<>
QSize arg( const char* arg )
    {
    const char *s = arg;
    char *end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int w = strtol(s, &end, 10);
    s = (const char *)end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int h = strtol(s, &end, 10);
    return QSize( w, h );
    }

template<>
QRect arg( const char* arg )
    {
    const char *s = arg;
    char *end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int p1 = strtol(s, &end, 10);
    s = (const char *)end;
    bool legacy = (*s == 'x');
    while(*s && !isdigit(*s) && *s != '-') s++;
    int p2 = strtol(s, &end, 10);
    s = (const char *)end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int p3 = strtol(s, &end, 10);
    s = (const char *)end;
    while(*s && !isdigit(*s) && *s != '-') s++;
    int p4 = strtol(s, &end, 10);
    if (legacy)
        {
        return QRect( p3, p4, p1, p2 );
        }
    return QRect( p1, p2, p3, p4 );
    }

template<>
int arg( const char* arg )
    {
    return atoi( arg );
    }

template<>
long arg( const char* arg )
    {
    return atol( arg );
    }

template<>
QString arg( const char* arg )
    {
    return QString::fromLocal8Bit( arg );
    }

// all toPrint() are based on dcop code, to make it compatible with KDE3
template< typename T >
QString toPrint( T arg );

template<>
QString toPrint( bool arg )
    {
    return arg ? "true" : "false";
    }

template<>
QString toPrint( int arg )
    {
    return QString::number( arg );
    }

template<>
QString toPrint( long arg )
    {
    return QString::number( arg );
    }

template<>
QString toPrint( QString arg )
    {
    return arg.toLocal8Bit();
    }

template<>
QString toPrint( QPoint arg )
    {
    return QString().sprintf( "+%d+%d", arg.x(), arg.y() );
    }

template<>
QString toPrint( QSize arg )
    {
    return QString().sprintf( "%dx%d", arg.width(), arg.height() );
    }

template<>
QString toPrint( QRect arg )
    {
    return QString().sprintf( "%dx%d+%d+%d", arg.width(), arg.height(), arg.x(), arg.y() );
    }

template< typename T >
QString toPrint( QList< T > arg )
    {
    QString result;
    for( int i = 0;
         i < arg.count();
         ++i )
        {
        result += toPrint( arg[ i ] );
        if( i != arg.count() - 1 )
            result += '\n';
        }
    return result;
    }

template< typename T >
static void print( T arg )
    {
    QString txt = toPrint( arg );
    if( !txt.isEmpty())
        QTextStream( stdout ) << txt << endl;
    }

template< typename R >
static void callFunction( R ( *function )(), int argc, const char* [] )
    {
    if( argc != 2 )
        usage();
    print( function());
    }

template<>
void callFunction( void ( *function )(), int argc, const char* [] )
    {
    if( argc != 2 )
        usage();
    function();
    }

template< typename R, typename T >
static void callFunction( R ( *function )( T t ), int argc, const char* argv[] )
    {
    if( argc != 3 )
        usage();
    print( function( arg< T >( argv[ 2 ] )));
    }

template< typename T >
static void callFunction( void ( *function )( T t ), int argc, const char* argv[] )
    {
    if( argc != 3 )
        usage();
    function( arg< T >( argv[ 2 ] ));
    }

template< typename R, typename T1, typename T2 >
static void callFunction( R ( *function )( T1 t1, T2 t2 ), int argc, const char* argv[] )
    {
    if( argc != 4 )
        usage();
    print( function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] )));
    }

template< typename T1, typename T2 >
static void callFunction( void ( *function )( T1 t1, T2 t2 ), int argc, const char* argv[] )
    {
    if( argc != 4 )
        usage();
    function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ));
    }

template< typename R, typename T1, typename T2, typename T3 >
static void callFunction( R ( *function )( T1 t1, T2 t2, T3 t3 ), int argc, const char* argv[] )
    {
    if( argc != 5 )
        usage();
    print( function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] )));
    }

template< typename T1, typename T2, typename T3 >
static void callFunction( void ( *function )( T1 t1, T2 t2, T3 t3 ), int argc, const char* argv[] )
    {
    if( argc != 5 )
        usage();
    function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ));
    }

template< typename R, typename T1, typename T2, typename T3, typename T4 >
static void callFunction( R ( *function )( T1 t1, T2 t2, T3 t3, T4 t4 ),
    int argc, const char* argv[] )
    {
    if( argc != 6 )
        usage();
    print( function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ),
        arg< T4 >( argv[ 5 ] )));
    }

template< typename T1, typename T2, typename T3, typename T4 >
static void callFunction( void ( *function )( T1 t1, T2 t2, T3 t3, T4 t4 ),
    int argc, const char* argv[] )
    {
    if( argc != 6 )
        usage();
    function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ),
        arg< T4 >( argv[ 5 ] ));
    }

template< typename R, typename T1, typename T2, typename T3, typename T4, typename T5 >
static void callFunction( R ( *function )( T1 t1, T2 t2, T3 t3, T4 t4, T5 t5 ),
    int argc, const char* argv[] )
    {
    if( argc != 7 )
        usage();
    print( function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ),
        arg< T4 >( argv[ 5 ] ), arg< T5 >( argv[ 6 ] )));
    }

template< typename T1, typename T2, typename T3, typename T4, typename T5 >
static void callFunction( void ( *function )( T1 t1, T2 t2, T3 t3, T4 t4, T5 t5 ),
    int argc, const char* argv[] )
    {
    if( argc != 7 )
        usage();
    function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ),
        arg< T4 >( argv[ 5 ] ), arg< T5 >( argv[ 6 ] ));
    }

template< typename R, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6 >
static void callFunction( R ( *function )( T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6 ),
    int argc, const char* argv[] )
    {
    if( argc != 8 )
        usage();
    print( function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ),
        arg< T4 >( argv[ 5 ] ), arg< T5 >( argv[ 6 ] ), arg< T6 >( argv[ 7 ] )));
    }

template< typename T1, typename T2, typename T3, typename T4, typename T5, typename T6 >
static void callFunction( void ( *function )( T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6 ),
    int argc, const char* argv[] )
    {
    if( argc != 8 )
        usage();
    function( arg< T1 >( argv[ 2 ] ), arg< T2 >( argv[ 3 ] ), arg< T3 >( argv[ 4 ] ),
        arg< T4 >( argv[ 5 ] ), arg< T5 >( argv[ 6 ] ), arg< T6 >( argv[ 7 ] ));
    }

template< typename T >
static void tryCallFunction( T function, const char* fname, int argc, const char* argv[] )
    {
    if( qstricmp( argv[ 1 ], fname ) != 0 )
        return;
    callFunction( function, argc, argv );
    XCloseDisplay( display());
    exit( 0 );
    }

#define tryFunction( function ) \
    tryCallFunction( function, #function, argc, const_cast< const char** >( argv ))

void dispatch( int argc, char* argv[] )
    {
    if( argc < 2 )
        usage();
    tryFunction( version );
    // they are in stacking order, from the topmost
    tryFunction( normalWindows );
    tryFunction( allWindows );
    // when empty, match all; without <2>
    tryFunction( findNormalWindows );
    tryFunction( findAllWindows );
    tryFunction( currentDesktop );
    tryFunction( setCurrentDesktop );
    tryFunction( numberOfDesktops );
    tryFunction( activeWindow );
    tryFunction( forceActiveWindow );
    // without <2>, case-insensitive
    tryFunction( caption );
    tryFunction( windowClass );
    tryFunction( windowRole );
    tryFunction( windowHostname );
    tryFunction( pid );
    // returns 0 for error, -1 for on all desktops
    tryFunction( windowDesktop );
    tryFunction( setWindowDesktop );
    // returns frame geometry
    tryFunction( frameGeometry );
    tryFunction( windowGeometry );
    tryFunction( setFrameGeometry );
    tryFunction( setWindowGeometry );
    tryFunction( resizeWindow );
    tryFunction( windowSize );
    tryFunction( frameSize );
    tryFunction( moveFrame );
    tryFunction( framePosition );
    tryFunction( windowPosition );
    tryFunction( fullScreenWindow );
    tryFunction( windowFullScreen );
    tryFunction( keepAbove );
    tryFunction( keptAbove );
    tryFunction( keepBelow );
    tryFunction( keptBelow );
    tryFunction( shadeWindow );
    tryFunction( windowShaded );
    tryFunction( maximize );
    // 0 - no, 1 - horiz, 2 - vert, 3 - full
    tryFunction( windowMaximized );
    // does not activate the window
    tryFunction( minimize );
    tryFunction( unminimize );
    tryFunction( minimized );
    tryFunction( raiseWindow );
    tryFunction( lowerWindow );
    tryFunction( closeWindow );

    usage();
    }
