/*****************************************************************

Copyright (C) 2006,2009 Lubos Lunak <l.lunak@kde.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#include "netwm.h"
#include "kwindowsystem.h"
#include "kwindowinfo.h"

const int SUPPORTED_WINDOW_TYPES_MASK = NET::NormalMask | NET::DesktopMask | NET::DockMask
    | NET::ToolbarMask | NET::MenuMask | NET::DialogMask /*| NET::OverrideMask*/ | NET::TopMenuMask
    | NET::UtilityMask | NET::SplashMask;

static QList< long > findWindows( QString caption, QString wclass, QString role,
    QString hostname, int pid, bool currentDesktop, bool normal )
    {
    NETRootInfo info( display(), NET::ClientListStacking );
    QList< long > ret;
    long mask1 = 0;
    long mask2 = 0;
    if( !caption.isEmpty())
        mask1 |= NET::WMName;
    if( !wclass.isEmpty())
        mask2 |= NET::WM2WindowClass;
    if( !role.isEmpty())
        mask2 |= NET::WM2WindowRole;
    if( !hostname.isEmpty())
        mask2 |= NET::WM2ClientMachine;
    if( currentDesktop )
        mask1 |= NET::WMDesktop;
    if( normal )
        mask1 |= NET::WMWindowType;
    for( int i = info.clientListStackingCount() - 1;
         i >= 0;
         --i )
        {
        WId w = info.clientListStacking()[ i ];
        KWindowInfo winfo( w, mask1, mask2 );
        if( !winfo.valid())
            continue;
        if( !caption.isEmpty() && !winfo.name().contains( caption, Qt::CaseInsensitive ))
            continue;
        if( !wclass.isEmpty()
            && !QString::fromUtf8( winfo.windowClassName() + ' ' + winfo.windowClassClass())
                .contains( wclass, Qt::CaseInsensitive ))
            {
            continue;
            }
        if( !role.isEmpty() && !QString::fromUtf8( winfo.windowRole())
            .contains( role, Qt::CaseInsensitive ))
            {
            continue;
            }
        if( !hostname.isEmpty() && !QString::fromUtf8( winfo.clientMachine())
            .contains( hostname, Qt::CaseInsensitive ))
            {
            continue;
            }
        if( currentDesktop && !winfo.isOnCurrentDesktop())
            continue;
        if( normal )
            {
            NET::WindowType type = winfo.windowType( SUPPORTED_WINDOW_TYPES_MASK );
            if( type != NET::Unknown && type != NET::Normal )
                continue;
            }
        if( pid != 0 )
            {
            NETWinInfo winfo2( display(), w, rootWindow(), NET::WMPid );
            if( winfo2.pid() != pid )
                continue;
            }
        ret.append( w );
        }
    return ret;
    }

static QList< long > normalWindows( bool currentDesktop )
    {
    return findWindows( QString(), QString(), QString(), QString(), 0, currentDesktop, true );
    }

static QList< long > allWindows( bool currentDesktop )
    {
    return findWindows( QString(), QString(), QString(), QString(), 0, currentDesktop, false );
    }

static QList< long > findNormalWindows( QString caption, QString wclass, QString role,
    QString hostname, int pid, bool currentDesktop )
    {
    return findWindows( caption, wclass, role, hostname, pid, currentDesktop, true );
    }

static QList< long > findAllWindows( QString caption, QString wclass, QString role,
    QString hostname, int pid, bool currentDesktop )
    {
    return findWindows( caption, wclass, role, hostname, pid, currentDesktop, false );
    }

static int currentDesktop()
    {
    return KWindowSystem::currentDesktop();
    }

static void setCurrentDesktop( int desktop )
    {
    KWindowSystem::setCurrentDesktop( desktop );
    }

static int numberOfDesktops()
    {
    return KWindowSystem::numberOfDesktops();
    }

static long activeWindow()
    {
    NETRootInfo info( display(), NET::ActiveWindow );
    return info.activeWindow();
    }

static void forceActiveWindow( long window )
    {
    KWindowSystem::forceActiveWindow( window );
    }

static QString caption( long window )
    {
    return KWindowInfo( window, NET::WMName ).name();
    }

static QString windowClass( long window )
    {
    KWindowInfo winfo( window, 0, NET::WM2WindowClass );
    if( !winfo.valid())
        return QString();
    return winfo.windowClassName() + ' ' + winfo.windowClassClass();
    }

static QString windowRole( long window )
    {
    return KWindowInfo( window, 0, NET::WM2WindowRole ).windowRole();
    }

static QString windowHostname( long window )
    {
    return KWindowInfo( window, 0, NET::WM2ClientMachine ).clientMachine();
    }

static int pid( long window )
    {
    NETWinInfo winfo( display(), window, rootWindow(), NET::WMPid );
    return winfo.pid();
    }

static int windowDesktop( long window )
    {
    KWindowInfo winfo( window, NET::WMDesktop );
    if( !winfo.valid())
        return 0;
    return winfo.desktop();
    }

static void setWindowDesktop( long window, int desktop )
    {
    KWindowSystem::setOnDesktop( window, desktop );
    }

static QRect frameGeometry( long window )
    {
    return KWindowInfo( window, NET::WMFrameExtents ).frameGeometry();
    }

static QRect windowGeometry( long window )
    {
    return KWindowInfo( window, (unsigned)NET::WMGeometry ).geometry();
    }

static void setFrameGeometry( long window, int x, int y, int width, int height )
    {
    KWindowInfo winfo( window, (unsigned)NET::WMGeometry | NET::WMFrameExtents );
    NETRootInfo info( display(), 0 );
    // calculate window size from given frame size
    width -= winfo.frameGeometry().width() - winfo.geometry().width();
    height -= winfo.frameGeometry().height() - winfo.geometry().height();
    // NorthWest gravity, x,y,w,h specified, source is tool
    info.moveResizeWindowRequest( window, 1 | ( 0x0f << 8 ) | ( 2 << 12 ), x, y, width, height );
    }

static void setWindowGeometry( long window, int x, int y, int width, int height )
    {
    NETRootInfo info( display(), 0 );
    // Static gravity, x,y,w,h specified, source is tool
    info.moveResizeWindowRequest( window, 10 | ( 0x0f << 8 ) | ( 2 << 12 ), x, y, width, height );
    }

static void resizeWindow( long window, int width, int height )
    {
    NETRootInfo info( display(), 0 );
    // window's gravity, w,h specified, source is tool
    info.moveResizeWindowRequest( window, 0 | ( 1 << 10 ) | ( 1 << 11 ) | ( 2 << 12 ), 0, 0, width, height );
    }

static QSize frameSize( long window )
    {
    return KWindowInfo( window, NET::WMFrameExtents ).frameGeometry().size();
    }

static QSize windowSize( long window )
    {
    return KWindowInfo( window, (unsigned)NET::WMGeometry ).geometry().size();
    }

static void moveFrame( long window, int x, int y )
    {
    NETRootInfo info( display(), 0 );
    // NorthWest gravity, x,y specified, source is tool
    info.moveResizeWindowRequest( window, 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), x, y, 0, 0 );
    }

static QPoint framePosition( long window )
    {
    return KWindowInfo( window, NET::WMFrameExtents ).frameGeometry().topLeft();
    }

static QPoint windowPosition( long window )
    {
    return KWindowInfo( window, (unsigned)NET::WMGeometry ).geometry().topLeft();
    }

static void changeState( long window, long state, bool set )
    {
    if( set )
        KWindowSystem::setState( window, state );
    else
        KWindowSystem::clearState( window, state );
    }

static void fullScreenWindow( long window, bool fs )
    {
    changeState( window, NET::FullScreen, fs );
    }

static bool windowFullScreen( long window )
    {
    return KWindowInfo( window, NET::WMState ).hasState( NET::FullScreen );
    }

static void keepAbove( long window, bool keep )
    {
    changeState( window, NET::KeepAbove, keep );
    }

static bool keptAbove( long window )
    {
    return KWindowInfo( window, NET::WMState ).hasState( NET::KeepAbove );
    }

static void keepBelow( long window, bool keep )
    {
    changeState( window, NET::KeepBelow, keep );
    }

static bool keptBelow( long window )
    {
    return KWindowInfo( window, NET::WMState ).hasState( NET::KeepBelow );
    }

static void shadeWindow( long window, bool shade )
    {
    changeState( window, NET::Shaded, shade );
    }

static bool windowShaded( long window )
    {
    return KWindowInfo( window, NET::WMState ).hasState( NET::Shaded );
    }

static void maximize( long window, bool horiz, bool vert )
    {
    if( horiz || vert )
        changeState( window, ( horiz ? NET::MaxHoriz : 0 ) | ( vert ? NET::MaxVert : 0 ), true );
    if( !horiz || !vert )
        changeState( window, ( !horiz ? NET::MaxHoriz : 0 ) | ( !vert ? NET::MaxVert : 0 ), false );
    }

static int windowMaximized( long window )
    {
    switch( KWindowInfo( window, NET::WMState ).state() & ( NET::Max ))
        {
        case NET::MaxHoriz:
            return 1;
        case NET::MaxVert:
            return 2;
        case NET::Max:
            return 3;
        case 0:
        default:
            return 0;
        }
    }

static void minimize( long window )
    {
    KWindowSystem::minimizeWindow( window );
    }

static void unminimize( long window )
    {
    KWindowSystem::unminimizeWindow( window );
    }

static bool minimized( long window )
    {
    return KWindowInfo( window, NET::WMState | NET::XAWMState ).isMinimized();
    }

static void raiseWindow( long window )
    {
    KWindowSystem::raiseWindow( window );
    }

static void lowerWindow( long window )
    {
    KWindowSystem::lowerWindow( window );
    }

static void closeWindow( long window )
    {
    NETRootInfo info( display(), 0 );
    info.closeWindowRequest( window );
    }
